' PARSERWORDS.BAS

' A Very Fast Configurable Parser (PB Assembler)
' with Test Demo
'
' Charles E V Pegge
' 19 August 2006
' updated 23 August 2006
' charles@pegge.net
' www.pegge.net
'
' with thanks to Paul Dixon, Michael Mattias, Eros Olmi and other members of the PowerBasic
' Peer Support Forums.

' A text file of the Bible, used in this demo may be obtained from:
' http://www.talsystems.com/contest12/contest12.zip

'
' Parsing rules using the default TokenType table:
'
' space, and all lower ascii and all non alphanumerics are treated as delimiters. eg:
' But an underscore _ (ascii 95) is treated as an alphanumeric eg:
'
'   FREE.LUNCH is interpreted as three words
'   FREE_LUNCH is one word

' all other non alphanumerics including '-' (minus) are self delimiting eg:
'
'   + - . ,  { } are always recognised as individual words even when unspaced
'
'
' But quoted text is treated as an entire word including the outer quotes.
' inner quotes can be embedded eg:
'
'    "She said 'HI' "
'    'She said "HI" '
'
' this Demo requires bible12.txt is placed in the
' same folder as this ParseTableTest.exe.
'
'
' quick Ascii code reference:
' ---------------------------
'  various control chars 0-31:
'  null      0
'  htab      9
'  linefeed  10
'  vtab      11
'  formfeed  12
'  creturn   13
'  escape    27
'  space     32
'  !"#$%&'   33-39
'  ()*+,-./  40-47
'  0-9       48-57
'  :;<=>?@   58-64
'  A-Z       65-90
'  [\]^_`    91-96
'  a-z       97-122
'  {|}~      123-126
' ---------------------------



#COMPILE EXE
#DIM ALL
#INCLUDE "WIN32API.INC"

MACRO tabs=+CHR$(9)+
MACRO newline=+$CRLF+ _

GLOBAL dummy AS LONG

#INCLUDE "ParserWords.inc"

FUNCTION WINMAIN (BYVAL hInstance AS DWORD,BYVAL hPrevInstance AS DWORD,BYVAL lpCmdLine AS ASCIIZ PTR,BYVAL iCmdShow AS LONG) AS LONG


  %MaxEntries=2000  ' an arbitrary figure
'
DIM cmz AS STRING
'sample Reserved Word list.
'
' any word found on this list will return
' the corresponding 2 byte hexadecimal code
' instead of the TokenType Code from the table below.

' For demo purposes we will use this list to scan the Bible
' and provide a word count.

' Keyword List Rules:

' The token must always be a 4 character hexadecimal number.
' There must be only one space (or tab) between the HexCode and the keyword.
' Keywords must start with a minimum of 2 alphas followed by any
' alphas or numbers from 0 to a maximum of 255 additional characters.
' Thus a keyword like V3 is invalid but TAN2 is okay.
' The keywords ore not case sensitive.
' Apart from the single space between code and keyword, a minimum of 1 spacing character
' is used to separate the data items.
' Spacing characters are any from ascii 0 to ascii 32 inclusive, including
' CRs, LFs TABs etc.
'
cmz=_
" 0230 verily"+_
" 0231 the"+_
" 0232 thee"+_
" 0233 unto"+_
" 0234 peace"+_
" 0235 pasture"+_
" 0236 lamb"+_
" 0237 flock"+_
" 0238 wrath"+_
" 0239 fire"+_
" 023A brimstone"+_
" 023B heaven"+_
" 023C earth"+_
" "

' TokenType Lookup Table

 DIM tbl AS GLOBAL STRING
 tbl=NUL$(&h2100+LEN(cmz)+%MaxEntries)

 'the TokenType Table:

 '0123456789ABCDEF'     '   codings
 MID$(tbl,1)= _                 '
 "             $  " + _ '0  space/ignore      32
 "                " + _ '1  quote marks     ! 33
 " #!####!########" + _ '2  symbols         # 35 min 34
 "0000000000######" + _ '3  numbers         0 48
 "#AAAAAAAAAAAAAAA" + _ '4  alpha uppercase A 65
 "AAAAAAAAAAA####_" + _ '5  alpha uppercase A 65
 "#aaaaaaaaaaaaaaa" + _ '6  alpha lowercase a 97
 "aaaaaaaaaaa#####" + _ '7  alpha lowercase a 97
 "hhhhhhhhhhhhhhhh" + _ '8  HIGHER ASCII    h 104
 "hhhhhhhhhhhhhhhh" + _ '9
 "hhhhhhhhhhhhhhhh" + _ 'A
 "hhhhhhhhhhhhhhhh" + _ 'B
 "hhhhhhhhhhhhhhhh" + _ 'C
 "hhhhhhhhhhhhhhhh" + _ 'D
 "hhhhhhhhhhhhhhhh" + _ 'E
 "hhhhhhhhhhhhhhhh" + _ 'F
 ""
'
DIM tblp AS GLOBAL BYTE PTR: tblp=STRPTR(tbl)
'
' TokenType Table Rules
'
' These 4 rules are hard coded into the Assembler parsing routine:
'
' 1. any code less than 33 is treated as space.
' 2. code 33 signifies a quote, and all characters between and including the matcing quote marks
'    will be included in the word
' 3. any code  from 34 to 47 is treated as self-delimiting symbol (or punctuation).
' 4. any code  greater than 47 is treated as text (so that includes the underscore 95).
'
'
' Configurability
'
' within these rules you can devise your own codings.
' When the GetNextToken is called, it will return the coding associated with the first character of the word.
' It is possible to generate codes for 'return' (13), and 'tab' (9) or any other lower ascii
' if your lexicon requires it. In the above table, a return character (13) will generate a $ (36).
' You can use similar methods for catching Escape codes and other embedded binaries,
' You can designate any character to be a quote marker as well as the usual " (34) and ' (39), by
' placing a ! (33) in the table at the marker's ascii position.

' If you do not want line sensitivity then replace the $ with a space at position 13 in this table.
'

DIM errp AS LONG, ii AS LONG, jj AS LONG

ReadInWords(cmz)
' this function returns several items of information on the CPU registers.
' whereas a normal function only returns things in its eax register.
' These items can be stored directly from the registers into LONG variables,
' immediately following the function call.
! mov errp,eax
! mov ii,edx
! mov jj,ecx
' This enables error reporting, if anything went wrong reading the reserved words.
IF errp THEN
 errp=MAX(errp-STRPTR(cmz)-8,1)
 MSGBOX "Problem with tokenlist about here: "+MID$(cmz,errp,16) newline
 HEX$(STRPTR(cmz))+" "+HEX$(ii)+" ecx: "+HEX$(jj)
 EXIT FUNCTION
END IF

'
DIM dString    AS STRING              ' the text to be read
DIM lString    AS LONG                ' length of this text
DIM pString    AS BYTE PTR            ' actual address of the start of text
DIM TokenStart AS LONG: TokenStart=1  ' string index to the 1st character of a word. (this is an output-only value)
DIM TokenEnd  AS LONG                 ' string index to the position immediately following the word. (set this to wherever you want to begin)
DIM TokenType(1024) AS LONG           ' used for counting words by coding-type
'
DIM re AS LONG                        ' the number of repeat reads to be done in the test
'
' sample text

'sample with slighty ammended lexical rules
ERRCLEAR
OPEN "bible12.txt" FOR INPUT AS 1:CLOSE 1
IF ERR THEN EXIT FUNCTION
OPEN "bible12.txt" FOR BINARY ACCESS READ AS 1:GET$ 1,LOF(1),dString:CLOSE 1

' adaptations for the Bible test:
' double quotes are treated as normal punctuation.
' apostrophes (39) are treated as part of word and defined as type 95, like an underscore.
' These codings are patched into the table: pointers do this for us nicely.
@tblp[34]=35 ' # ' double quote to be designated as symbol (or punctuation).
@tblp[39]=95 ' _ ' single quote alias apostrophe to be treated like an underscore.
'
re=1: GOSUB test_and_display: ' as this is a very large text, just read it once.

EXIT FUNCTION

test_and_display:

LOCAL i AS LONG    ' iterator for repeated reads of the text
LOCAL j AS LONG    ' word counter
LOCAL t AS DWORD   ' for measuring lapsed time in millisecs
LOCAL ot AS DWORD  ' records the lapsed time on a dummy run
                   ' this is used to measure the computing time not spent doing Assembler code

lString=LEN(dString)
pString=STRPTR(dString)  ' address of the string content
                         '
                         ' first a dummy loop to measure the PowerBasic overheads
                         '
DIM s AS STRING          ' string to receive the report to be displayed
DIM dummy AS GLOBAL LONG ' a flag to exclude Assembler code execution
'
'do a dummy run to measure PB and system overheads
'
s=LEFT$(dString,60):dummy=1:GOSUB reading_test
ot=t      ' record the overheads time
SLEEP 100 ' catchup with housekeeping. Seems to give more stable time lapse figure.
          ' now do the real thing
s=LEFT$(dString,60)+"..."+$CRLF:dummy=0:GOSUB reading_test
'
' Report showing word counts for reserved words as well as other stats.
'
MSGBOX s                                  newline
 ""                                       newline
 STR$(TokenType(&h0230)) tabs "verily"    newline
 STR$(TokenType(&h0231)) tabs "the"       newline
 STR$(TokenType(&h0232)) tabs "thee"      newline
 STR$(TokenType(&h0233)) tabs "unto"      newline
 STR$(TokenType(&h0234)) tabs "peace"     newline
 STR$(TokenType(&h0235)) tabs "pasture"   newline
 STR$(TokenType(&h0236)) tabs "lamb"      newline
 STR$(TokenType(&h0237)) tabs "flock"     newline
 STR$(TokenType(&h0238)) tabs "wrath"     newline
 STR$(TokenType(&h0239)) tabs "fire"      newline
 STR$(TokenType(&h023a)) tabs "brimstone" newline
 STR$(TokenType(&h023b)) tabs "heaven"    newline
 STR$(TokenType(&h023c)) tabs "earth"     newline
 "" newline
 STR$(lString)       tabs "bytes"                   newline
 "" newline
 STR$(TokenType(48)) tabs "numbers"                 newline
 STR$(TokenType(97)) tabs "other lowercase words"   newline
 STR$(TokenType(65)) tabs "other proper words"      newline
 STR$(TokenType(35)) tabs "punctuation"             newline
 STR$(TokenType(36)) tabs "lines"                   newline
 STR$(re*j)          tabs "total items"             newline
 "" newline
 STR$(INT(t-ot))     tabs "milliseconds ASM"        newline
 STR$(INT(ot))       tabs "milliseconds Powerbasic" newline
 STR$(INT(t))        tabs "milliseconds total time" newline
 "" newline
 STR$(INT(j/t))      tabs "words/millisec" +$CRLF   newline
 MID$(dString,TokenStart-7,8)                       newline
 ""
 '
RETURN
'
reading_test:

DIM ShowWords AS LONG: ShowWords=1
DIM priority AS LONG
FOR i=0 TO 1023:tokentype(i)=0:NEXT            ' zero all token counts
j=0                                           ' zero the word counter
priority=GetPriorityClass(GetCurrentProcess)  ' save current thread priority
i=SetPriorityClass(GetCurrentProcess,&h00000100)   ' set Priority to REAL TIME

IF i=0 THEN MSGBOX "unable to get RealTime priority for accurate measurement"

t=getTickCount()                              ' get the current millisecond count since last boot
                                              ' but if you leave your computer on for 49.7 days then
                                              ' this counter will turnover!
LOCAL vv AS LONG
LOCAL tl AS LONG
'
FOR i=1 TO re
 TokenStart=0
 TokenEnd=1
 WHILE TokenEnd<lString

  '  every time you invoke GetNextToken it will get the next word until the end of the string
  '  is reached, then TokenEnd and Token start will be greater than the dString length.
  '  The return value is the token type matching the 1st character of the word,
  '  or if the word has been identified in the list, its own token will be returned.
  '  Here, we use this value to count different tokens or tokentypes as they occur.
  '  The counts are reported in the message box above.
  '
  vv=GetNextToken(STRPTR(dString), lString, TokenStart, TokenEnd)
  IF (vv>0)AND(vv<1024) THEN INCR TokenType(vv)
  INCR j
 WEND
NEXT
t=getTickCount()-t 'record lapsed time in milliseconds
SetPriorityClass(GetCurrentProcess(),priority) ' restore priority: this is usually &h00000020 NORMAL PRIORITY
RETURN

END FUNCTION
