' A Very Fast Configurable Parser (PB Assembler)

' For use with ParserWords.bas which provides data framework and demo.

'
' Charles E V Pegge
' 19 August 2006
' updated 23 August 2006
' charles@pegge.net
' www.pegge.net
'
' with thanks to Paul Dixon, Michael Mattias, Eros Olmi and other members of the PowerBasic
' Peer Support Forums.


FUNCTION ReadInWords(cmz AS STRING) AS LONG
#REGISTER NONE
' procedure to construct lookup tables for reserved words and their tokens

'======================='
! push ebp              ' save registers used by PB which we want to use
'-----------------------'
! mov ecx,tblp          ' address of the codings table
'                       ' we assume no further use of the ebp reg BASIC var loading
! mov ebp,cmz           ' address of string handle
'-----------------------'
!                       ' index registers can now be overwritten safely
!                       ' because we have loaded all the variables we need for now.
! mov ebp,[ebp]         ' address of string content
! mov edx,[ebp-4]       ' length of string to the left of the contant
! add edx,ebp           ' add its address+length to obtain the its beyond-end pointer
! mov ebx,ecx           ' move base address of tokentype table (assumed to be 256 bytes long)
! add ebx,&h100         ' base address for lookup table which sits immediately above.
! mov edi,ebx           ' the lookup table works out like this:
!                       ' each entry location is derived from the first 2 letters of a word
!                       ' there are 32x32=1024 entries with a stride of 8 giving an overall
!                       ' length of 8192 bytes (&h2000).
!                       ' each entry contains 2 dwords:
!                       ' #1: address of 1st entry in data list or null if there is none.
!                       ' #2: address of last data link position in this chain
!                       ' the string data list sits on top of this lookup table
! add edi,&h2000        ' so this is next available address in data chain address
'-----------------------'
next_sp:                ' look at another char
'-----------------------'
! inc ebp               '
! cmp edx,ebp           ' check if we have reached the end of string
! jle exit_s            ' finish because there is no keyword following
! mov al,[ebp]          ' load char
! cmp al,32             ' is it a space or less?
! jle short next_sp     ' if yes then try another char
!                       ' drop thru if not a space
'-----------------------'
! add ebp,4             ' skip 4 bytes. The next byte is assumed to be a space.
! cmp edx,ebp           ' check if we have reached the end of string
! jle short exit_s      ' finish because there is no keyword following
! mov ecx,[ebp-4]       ' assume this is 4 byte token data
!                       '
! call token_encode     ' convert ascii to token code to binary form
!                       ' the result is held in ecx after conversion
!                       ' error check:
! cmp byte ptr [ebp],&h20 ' the byte should always be a space or less
! jg short exit_s_err   ' there is a misalignment somewhere so quit
'-----------------------'
next_ch:                '
'-----------------------'
! add ebp,3             '
! cmp edx,ebp           ' have we exceeded the length of the string
! jle short exit_s_err  ' if so then this is a premature end so report an error
! sub ebp,2             '
! movzx eax,word ptr [ebp] ' we are going to look at the first 2 letters of the word
! and eax,&h00005f5f    ' first mask off the upper 2 bytes of eax and also upper ascii and lowercase
! sub eax,&h00004040    ' deduct 64 from each char so we end up with A=1 B=2 etc
!                       ' Alpha range: 01X00001 to 01X11111 where X is the lowercase bit.
!                       ' After -&h40: 00000001 to 00011111.
!                       ' this encodes 0 to 31 but we only use the alphabetic 1-26 for A-Z
!                       '
!                       '
! jl short exit_s_err   ' if the result is negative then it was not not alpha so exit
!                       ' numbers in the 1st 2 positions are not supported.
! shl al,3              ' multiply the lower byte by 8. but leave the upper byte alone.
!                       ' result of shift: 000XXXXX to XXXXX000
!                       ' combined lookup: 000XXXXX:XXXXXXXX range 0 to &h1FFF (0-8184) step 8
!                       ' we now have a 13 bit address offset for the lookup table.
!                       ' This table accommodates 2 dwords for each dual letter index
!                       ' the first dword will holds the address of the first record in the chain
!                       ' the second dword wikk hold the address where a link address is to be patched
!                       ' in whenever the chain is to be extended with another record
! mov esi,ebx           '
! add esi,eax           ' transfer offset to esi reg so eax is now free            '
! cmp dword ptr [esi],0 ' check the data at this location. Is it null?
!                       '
! jnz short go_addr     ' if it already contains an address then skip to go_adr
! mov [esi],edi         ' otherwise enter the address of the first data record in this chain
!                       '
go_addr:                '
!                       ' the second word will be filled in later when we have an end-of-chain adress
! push ebx              ' save ebx, we are going to need it.
! mov ebx,[esi+4]       ' get the address of the last link
! cmp ebx,0             ' if there isnt one (because this is the first word in the chain)
! jz short done_chain_link ' then skip the next instruction
! mov [ebx],edi         ' store the address of our next word in the chain
'-----------------------'
done_chain_link:        '
'-----------------------'
! pop ebx               ' restore ebx. so this holds the lookup table's base address again
! mov byte ptr [edi],0  ' default initial length of string-2.
!                       ' Because we already know the 1st 2 letters, only the rest of the word
!                       ' is used. This word tail may be 0 to 255 bytes long. It is
!                       ' followed by the 4 byte word token then the 4 byte link address
!                       ' to the next word
'-----------------------'
! push ecx              ' save the word token
! mov ecx,edi           ' note the start of the word string to use later when we
!                       ' determine the word length
! inc ebp               ' pre incr ebp because we ignore 1st 2 chars of word
!                       ' drop thru
'-----------------------'
next_wchar:             ' process the next letter in the word
'-----------------------'
! inc edi               ' increment word string ptr
! inc ebp               ' increment source pointer
! cmp edx,ebp           ' are we at the end of input string?
! jle short exit_wchar  ' then this is the end of the word so jump to next stage
! mov al,[ebp]          ' look at char
! cmp al,32             ' is it a space or less?
! jle short exit_wchar  ' past end of word
! and al,&h5f           ' convert uppercase also numbers 0-9 converted to &h10 to &h19
! mov [edi],al          ' store char to data string
! jmp short next_wchar  ' look at the next char
'-----------------------'
exit_wchar:               '
'-----------------------'
!                       '
! xchg edi,ecx          ' swap current pointer with the pointer to the word length
! sub ecx,edi           ' difference between start and end to give us the length of the word tail.
! dec ecx               ' deduct 1 to get the actual length of this tail
! mov [edi],cl          ' patch in this word tail length
! inc edi               ' reverse the edi movement to
! add edi,ecx           ' restore the edi position to beyond end of word
! pop ecx               ' finished using this register so restore token data
! mov [edi],ecx         ' store the token code immediately following the string
! add edi,4             ' advance edi to the chain link position
! mov [esi+4],edi       ' store this patch location for the data chain address
!                       ' this address is filled in when another word+token as added to the chain
! add edi,4             ' fwd edi to the next vacant position
! jmp next_sp           ' look at next input data
'-----------------------'
!                       '
exit_s:                 ' normal exit here
! mov ebp,0             ' null the pointer if all went well
!                       '
'-----------------------'
exit_s_err:             ' error states arrive here
!                       ' the location where the error occured is in ebp
!                       '
'-----------------------'
!                       ' transfer the following data to make it available on return
! mov eax,ebp           ' 1: eax now contains the address of any error, or null if things went ok
! mov edx,edi           ' 2: edx points beyond the end of the data record chain
!                       ' assume there is sufficient space to accommodate all the records
!                       ' as there is no limit checking on the data chain.
!                       ' the actual space required is length byte + token code + link address + wordlength -2
!                       ' the minimum length of the souce data is token code + space + wordlength + space
!                       ' so the stored data length is the length of the source data string + 1 byte extra for each word.
!                       ' Not too hard to estimate.
!                       '
!                       ' A clean table is assumed. Initialised with nulls. So if the table
!                       ' gets reused it must be reinitialised first.
!                       '
!                       ' to add further data, the next available position, held in edx
!                       ' must be recorded and used at the start of the repeated procedure.
'-----------------------'
 ! pop ebp              ' restored essential register for the system
'-----------------------'
! mov FUNCTION,eax      ' returns zero if ok otherwise it is the error's absulute address
!                       ' pointing into the source string.
EXIT FUNCTION           ' our only point of exit
'-----------------------'
!                       '
!                       '
'-----------------------'
token_encode:           ' Encode 4 byte token word into binary form
'-----------------------'
!                       ' this is a subroutine to be called
!                       ' without any params on the stack
!                       ' just the return address
!                       ' ascii token word is contained in ecx
!                       '
!                       ' result will be a hexadecimal to binary
!                       ' encoding in the lower half of ecx
!                       '
!                       ' other kinds of encoding possible by replacing
!                       ' this routine following a similar protocol
!                       '
! mov esi,edx           ' assume esi is not yet being used by caller
!                       ' so we save edx into it as
!                       ' this is quicker than pushing and popping
!                       ' eax assumed to be available
! xor eax,eax           ' set eax to zero
! xor edx,edx           ' set edx to zero
'-----------------------'
!                       ' for each byte
! and ecx,&h4f4f4f4f    ' knock out bits 10110000 to covert to uppercase
!                       ' and base decimal digits from zero
'-----------------------'
! mov al,ch             ' lowest byte
! cmp al,09             ' if greater than 9 then it as  one of ABCDEF
! jle short hex_adj1    '
! sub al,55             ' subtract 55 so that ascii A becomes hex A etc
hex_adj1:               '
! mov ah,cl             ' repeat this process using the next byte in ecx
! cmp ah,09             '
! jle short hex_adj2    '
! sub ah,55             '
hex_adj2:               '
! mov dh,al
! shr eax,4             ' shift the nibble in ah to the upper nible of al
! add dh,al             ' combine both upper and lower nibbles in dl
'-----------------------'
! shr ecx,16            ' shift ecx along by 2 bytes to process them
'-----------------------'
!                       ' repeat the process above, storing the result in dh
! mov al,ch             '
! cmp al,09             '
! jle short hex_adj3    '
! sub al,55             '
hex_adj3:               '
! mov ah,cl             '
! cmp ah,09             '
! jle short hex_adj4    '
! sub ah,55             '
hex_adj4:               '
! mov dl,al             '
! shr eax,4             ' bunch upper nibble
! add dl,al             ' combine
'-----------------------'
! mov ecx,edx           ' move the completed code into ecx ready for return
! mov edx,esi           ' restore edx
!                       ' esi and eax now contain trash but caller doesnt mind
! ret                   '
'======================='

END FUNCTION


FUNCTION GetNextToken(BYVAL pString AS BYTE PTR, BYVAL lString AS LONG, BYREF StartPos AS LONG , BYREF NextPos AS LONG) AS LONG
#REGISTER NONE
IF dummy THEN StartPos=NextPos+1: NextPos=NextPos+6: EXIT FUNCTION
                            '
'==========================='
!                           ' first save any registers used for Powerbasic variables
!                           ' in this instance, only the ebp register is essential since
!                           ' it is used to reference the parameters.
!                           ' everything else is already taken care of by the PB Function handling
!                           '
! push ebp                  ' save original base pointer
!                           '
! mov edx,pString           ' pointer to start of string
! dec edx                   ' offset by -1 for use with index
! mov esi,edx               ' hold this pointer for later use
! mov eax,edx               ' to find end of string
! add edx,lString           ' points to the last character in the string
! mov ebx,NextPos           ' getting parameter BYREF
! add eax,[ebx]             ' points to the character indexed by NextPos
!                           ' with basic variables installed,
!                           ' it is now safe to overwrite the base pointer register
! mov ebx,tblp              ' move the tbl pointer into ebx ready to do lookups
!                           ' we now lose the base pointer until we need it again to assign params
! mov ebp,eax               ' move the char pointer to the base pointer
! mov ecx,eax               ' default value for the start of word pointer
! xor edi,edi               ' clear edi, so it has a zero in it if you go past end of string              '
! jmp short test_space      ' Now the real work
'---------------------------'
next_space:                 ' to skip leading spaces tabs etc
'---------------------------'
! inc ebp                   ' advance the char pointer
! inc ecx                   ' advance tokenstart pointer
'---------------------------'
test_space:                 ' entry point into the next_space loop
'---------------------------'
! cmp edx,ebp               ' check if end of string
! jl short past_end         ' exit loop if end of string
! movzx eax, byte ptr [ebp] ' get character
!                           '
! movzx eax, byte ptr [ebx+eax] '
!                           '
!                           ' to look up its type code in tbl
! cmp al,32                 ' is it a space or less ?
! jle short next_space      ' if so then try next character
!                           '
'---------------------------' drop thru ..
start_of_word:              ' first non space char encountered
'---------------------------'
! mov edi,eax               ' save the coding
! cmp al,48                 ' if it is more than a symbol
! jg short try_id           ' then it is an alpha so try to identify it.
! jz short is_number        ' its a number
!                           '
! cmp al,33                 ' is it a quote type of symbol?
! jnz short end_of          ' if not then return a self-delimiting symbol
'---------------------------' drop thru for quotes
quote_string:               ' this is a quoted item, we need to find the end quote mark
'---------------------------'
                            '
! mov al,[ebp]              ' reload the specific quote mark
!                            ' drop thru
'---------------------------'
quote_string_traverse:      ' look for a matching quote mark
'---------------------------'
! inc ebp                   ' advance pointer for next char
! cmp edx,ebp               ' check if past the end of string
! jl short end_of           ' finish if so
! cmp al,[ebp]              ' check for matching quote mark
!                           '
! jnz short quote_string_traverse
!                           '
!                           ' if no match then try next
! jmp short end_of          ' we have our quote so finish
'---------------------------'
'---------------------------'
try_id:                     ' might be a reserved word
'---------------------------'
!                           '
! inc ebp                   ' check if beyond end of string
! cmp edx,ebp               ' is it past end of string
! jl short past_end         ' exit if past end
! movzx eax,word ptr [ebp-1]' look at 2 bytes to derive word index
! and eax,&h00005f5f        ' mask upper and convert to uppercase
! sub eax,&h00004040        '
! jl  short next_letter     ' reject numbers etc
!                           ' 5 bit+5 bit packed index with
!                           ' the first byte is bunched up 3 places for
! shl al,3                  ' stride of 8
!                           '
! add eax,256               ' the lookup table immediately follows the coding table
!                           ' the codings table is 256 bytes long and
!                           ' remember we are using a base1 index so offset is 257
! cmp dword ptr [ebx+eax],0 ' contains 0 ?
! jz short next_letter      ' if 0 then there is no word chain to check
!                           ' so revert to simple delimit checking
!                           ' otherwise edi is overwritten with the address of the
!                           ' data chain.
! mov edi,dword ptr [ebx+eax] ' assign address to edi if there is a potential
!                           ' match. edi is an absolute address to the word chain
'---------------------------'
is_number:                  ' entry point for numbers, bypassing id procedure
'---------------------------'
! inc ebp                   ' point to the third letter (or 2nd digit of number) and
!                           ' drop thru
'---------------------------'
next_letter:                ' continue scanning alphanumeric
'---------------------------'
!                           '
! cmp edx,ebp               ' past end of string?
! jl short word_identify    ' if so then go on to identify
!                           '
!                           '
! movzx eax,byte ptr [ebp]  ' get char
! mov al,[ebx+eax]          ' then lookup its type in tbl
! cmp al,&h30               ' if it is less than a number ?
! jl short word_identify    ' then delimited so go on to word identify
!                           '
! inc ebp                   ' advance char pointer
! jmp short next_letter     ' try another letter
'---------------------------'
'---------------------------'
end_of:                     ' results are ready to pass back to BASIC variables
'---------------------------'
!                           ' self delimiting symbols and end-quotes arrive here
! inc ebp                   ' increment char pointer
'---------------------------'
past_end:                   ' jump here when beyond end of string
'---------------------------'
!                           ' eax is our only scratch register
!                           '
! mov eax,ebp               ' hold the end position in eax before we pop the bp register
! sub ecx,esi               ' derive our start of word index
! sub eax,esi               ' derive the after-word index
!                           '
!                           ' it is now safe to restore any reserved registers
! pop ebp                   ' restore base pointer register
!                           ' we can now return results BYREF
! mov ebx,StartPos          '
! mov [ebx],ecx             '
! mov [ebx-4],eax           ' NextPos is next parameter along
!                           ' so it is now safe to store to any PowerBASIC variables
! mov FUNCTION,edi          ' store coding token type
                            ' if the string is null or contains no words or symbols
                            ' then StartPos will be greater than the length of the string
 EXIT FUNCTION              ' essential registers are popped from the stack
                            '
                            ' full identification
                            '
'==========================='

'==========================='
word_identify:              ' word identification
'---------------------------'
! cmp edi, 255              ' check the code
! jle short past_end        ' reject because edi only contains a tokentype,
!                           ' not an address to a data chain
! mov eax,ebp               ' check length
! sub eax,ecx               ' diff end and start pointers
! dec eax                   ' subtract another. is it zero or less ?
! jle short past_end        ' then its only only a single letter so finish now
!                           ' otherwise drop thru
'---------------------------'
! dec eax                   ' another -1 from eax gives the expected word tail length
!                           '
!                           ' save these registers so we can make use of them
! push esi                  '
! push edx                  '
! push ecx                  '
'---------------------------'
! push eax                  ' length of the word tail which we will use later
!                           ' reading it directly from the stack which is predecremented
!                           ' when you push something onto it, so we can read the length
!                           ' as [esp]. other values further up the stack can be accessed by
!                           ' [esp+4] [esp+8] [esp+12] etc
!                           '
! mov esi,ecx               ' pass the start pointer to esi reg
! inc esi                   ' point to the 2nd letter as this will be preincremented later
!                           ' so that we make our char matching from the 3rd letter
! mov edx,esi               ' holding register for start of compare. This is needed for
!                           ' resetting esi if and when we need to do another comparison.
!                           ' ecx is now available for use as an iterator.
!                           ' drop thru ..
'---------------------------'
next_string:                '
'---------------------------'
!                           '
! xor ecx,ecx               ' clear ecx, our iteration counter
! mov cl,[edi]              ' byte indicating word tail length
! cmp cl,[esp]              ' check if tail lengths match, the length of the keyword tail
!                           ' is held on the stack
!                           '
'---------------------------'
! jnz short no_length_match ' if they dont match then prepare to look at next
'---------------------------' or for left side compare only
'! jg short no_length_match ' if keyword greater then prepare to look at next
'---------------------------'
!                           '
! cmp cl,0                  ' check for zero length
! jz short found_string     ' its a 2 letter word, no tail to check, so skip the compare
!                           ' drop thru ..
'---------------------------'
compare_string:             '
'---------------------------'
!                           ' the first byte indicates how many bytes follow
!                           '
! inc esi                   ' next char in string
! inc edi                   ' next char in string table
! mov al,[esi]              ' load table char
! and al,&h5f               ' convert uppercase also numbers 0-9 converted to &h10 to &h19
! cmp al,[edi]              ' compare with string char
! jnz short no_match        ' if no match then exit loop
! dec ecx
! jnz compare_string       ' if ecx is zero then
!                           ' match perfect so drop thru
'---------------------------'
found_string:               '
'---------------------------'
!                           '
! inc edi                   ' increment table pointer so it points beyond
!                           ' the string pointer to the token dword
! mov edi,[edi]             ' get the token dword
!                           ' drop thru
'---------------------------'
done_word_identify:         ' exit point when successful
'---------------------------'
!                           '
!                           ' restore these registers
! pop eax                   ' eax is scratch but still needs to be popped.
! pop ecx                   '
! pop edx                   '
! pop esi                   '
! jmp short past_end        ' resume the exit procedure passing back to PB vars
'---------------------------'
'---------------------------'
no_length_match:            '
! inc edi                   ' go past the length byte                           '
no_match:                   '
'---------------------------'
! add edi,ecx               ' skip remaining char count of the word tail
! add edi,4                 ' skip the 4 byte token to get to the address of the next word.
! mov edi,[edi]             ' coukd be then address of next word or null
! cmp edi,0                 ' check if null, which means no more words in the chain to check.
! jz short word_id_fail     ' when edi=0 then failed to identify word -- ? restore coding in edi
! mov esi,edx               ' restore string pointer to its length byte
! jmp short next_string     ' try matching another table string
'---------------------------'
'---------------------------'
word_id_fail:               '
!                           ' restore the pointer regs we have been using
! pop eax                   '
! pop ecx                   '
! pop edx                   '
! pop esi                   '
!                           ' because there was no match for the word,
!                           ' we need to restore the original tokentype:
!                           '
! movzx eax, byte ptr [ecx] ' first get the first char of the word
! movzx edi, byte ptr [ebx+eax] ' then lookup its tokentype.
! jmp short past_end        ' resume the exit procedure passing back to PB vars
!                           '
'==========================='

END FUNCTION
